import socket

from .primatives import Primitive


class Connection:
    def __init__(self, ip_address: str = "localhost", port: int = 6000) -> None:
        self.__address = (ip_address, port)

        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 65535)

        self.__buffer_size = self.__socket.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)

    def __del__(self) -> None:
        self.__socket.close()

    def send_text(self, text: str, seconds: float = 0) -> None:
        self.__send(f'{{"text":{{"text":"{text}","seconds":{seconds}}}}}')

    def send_frame(self, primitives: list[Primitive], layer: int = 0) -> None:
        self.__send(f'{{"frame":{{"layer":{layer},"primitives":[{",".join([str(p) for p in primitives])}]}}}}')

    def __send(self, json: str) -> None:
        data = json.encode("ascii")

        if len(data) > self.__buffer_size:
            raise ValueError(f"The data size is {len(data)}, which exceeds the buffer size of {self.__buffer_size}.")

        self.__socket.sendto(data, self.__address)



import numpy as np

from ..link import Link
from ..matrix import Matrix

# Sets the height of the root so that the lowest part of the model is in
# contact with the floor.


def floor(root: Link) -> None:
    links = root.flatten()

    if not root.is_root:
        raise ValueError(f"{root.name} is not the root")

    joints_z = np.array([l.get_joint_world().z for l in links])

    ends_z = np.array([l.get_end_world().z for l in links])

    wheels_z = np.array([_wheel_lowest_point(l)[2] for l in links if l.get_wheel_axis_world()])

    min_z = np.min(np.concatenate((joints_z, ends_z, wheels_z)))

    root.joint = Matrix(xyz=root.joint.xyz - [0, 0, min_z], rotation=root.joint.rotation)


def _wheel_lowest_point(link: Link) -> np.ndarray:
    axis = link.get_wheel_axis_world().xyz

    lowest_spoke = np.cross(axis, np.cross(axis, np.array([0, 0, 1])))

    if np.isclose(lowest_spoke, 0).all():
        return link.get_joint_world().xyz

    lowest_spoke /= np.linalg.norm(lowest_spoke)

    return link.get_joint_world().xyz + (link.length * lowest_spoke)



from __future__ import annotations

import numpy as np

from .matrix import Matrix


class Link:
    def __init__(self, name, end: Matrix, wheel_axis: Matrix | None = None) -> None:
        self.__name = str(name)
        self.__origin = Matrix()  # link origin relative to the world (root origin is always coincident with world origin)
        self.__joint = Matrix()  # joint rotation relative to origin
        self.__end = end  # link end relative to origin
        self.__imu = Matrix(xyz=end.xyz / 2)  # IMU relative to origin
        self.__wheel_axis = Matrix(xyz=wheel_axis.xyz) if wheel_axis else None  # direction defined by wheel_axis.xyz
        self.__links: list[tuple[Link, Matrix]] = []  # [(link, matrix), ...] where matrix is the origin of the next link relative to the end of this link
        self.__is_root = True

        if wheel_axis and not np.isclose(np.dot(self.__end.xyz, self.__wheel_axis.xyz), 0):
            raise ValueError(f"Invalid values for {name}. Link end {self.__end.xyz} must be orthogonal to wheel_axis {self.__wheel_axis.xyz}.")

    @property
    def name(self) -> str:
        return self.__name

    @property
    def origin(self) -> Matrix:
        return self.__origin.copy()

    @property
    def joint(self) -> Matrix:
        return self.__joint.copy()

    @joint.setter
    def joint(self, joint: Matrix) -> None:
        if self.__is_root:
            self.__joint = joint
        else:
            self.__joint = Matrix(xyz=self.__joint.xyz, rotation=joint.rotation)  # ignore joint.xyz

        self.__update()

    @property
    def end(self) -> Matrix:
        return self.__end.copy()

    @property
    def imu(self) -> Matrix:
        return self.__imu.copy()

    @imu.setter
    def imu(self, imu: Matrix) -> None:
        self.__imu = Matrix(xyz=self.__imu.xyz, rotation=imu.rotation)  # ignore imu.xyz

    @property
    def wheel_axis(self) -> Matrix | None:
        return self.__wheel_axis.copy() if self.__wheel_axis else None

    @property
    def links(self) -> list[tuple[Link, Matrix]]:
        return self.__links

    @property
    def length(self) -> float:
        return np.linalg.norm(self.__end.xyz)

    @property
    def is_root(self) -> bool:
        return self.__is_root

    def __update(self, origin: Matrix | None = None) -> None:
        if origin is not None:
            self.__origin = origin

        for link, matrix in self.__links:
            link.__update(self.__origin * self.joint * self.__end * matrix)

    def connect(self, link: Link, matrix: Matrix = Matrix()) -> Link:  # matrix is the origin of the next link relative to the end of this link
        link.__is_root = False
        self.__links.append((link, matrix))
        self.__update()
        return self

    def get_joint_world(self) -> Matrix:
        return self.__origin * self.__joint

    def set_joint_world(self, joint_world: Matrix) -> None:
        self.joint = self.__origin.inverse * joint_world

    def get_end_world(self) -> Matrix:
        return self.__origin * self.__joint * self.__end

    def get_imu_world(self) -> Matrix:
        return self.__origin * self.__joint * self.__imu

    def set_imu_world(self, imu_world: Matrix) -> None:
        self.imu = self.__joint.inverse * self.__origin.inverse * imu_world

    def set_joint_from_imu_world(self, imu_world: Matrix) -> None:
        self.joint = self.__origin.inverse * imu_world * self.__imu.inverse

    def get_wheel_axis_world(self) -> Matrix | None:
        return Matrix(rotation=(self.__origin * self.__joint).rotation) * self.__wheel_axis if self.__wheel_axis else None

    def flatten(self) -> list[Link]:
        links = [self]

        for link, _ in self.__links:
            links += link.flatten()

        return links



from .link import Link
from .matrix import Matrix

# An interface to get and set Link joint matrix rotations as the three angles:
# alpha, beta, and gamma. The rotation follows the ZYX Euler sequence: alpha ->
# beta -> gamma. Alpha and gamma angles may range between +/-180 degrees; beta
# angles must not exceed or approach +/-90 degrees.
#
# The alignment of these angles should be chosen to best express the actual
# rotation of a given joint and to ensure that each angle does not exceed its
# permitted range.


class Joint:
    def __init__(
        self,
        link: Link,
        alignment: Matrix = Matrix(),  # if identity then alpha = rot_z, beta = rot_y, gamma = rot_x
        alpha_limit: tuple[float, float] | None = None,  # (<min>, <max>) or None if unlimited
        beta_limit: tuple[float, float] | None = None,  # (<min>, <max>) or None if unlimited
        gamma_limit: tuple[float, float] | None = None,  # (<min>, <max>) or None if unlimited
    ) -> None:
        self.__link = link
        self.__alignment = alignment
        self.__alpha_limit = alpha_limit
        self.__beta_limit = beta_limit
        self.__gamma_limit = gamma_limit

    @property
    def link(self) -> Link:
        return self.__link

    @property
    def alignment(self) -> Matrix:
        return self.__alignment.copy()

    @property
    def alpha_limit(self) -> tuple[float, float] | None:
        return self.__alpha_limit

    @property
    def beta_limit(self) -> tuple[float, float] | None:
        return self.__beta_limit

    @property
    def gamma_limit(self) -> tuple[float, float] | None:
        return self.__gamma_limit

    def get(self) -> tuple[float, float, float]:
        gamma, beta, alpha = (self.__alignment.inverse * self.__link.joint * self.__alignment).rot_xyz

        return alpha, beta, gamma

    def set(self, alpha: float = 0, beta: float = 0, gamma: float = 0) -> None:
        self.__link.joint = self.__alignment * Matrix(rot_x=gamma, rot_y=beta, rot_z=alpha) * self.__alignment.inverse



from ..link import Link
from ..matrix import Matrix

# Interpolates the joint of each link between the first and last linke of a
# contagious chain.


def interpolate(links: list[Link]) -> None:
    if len(links) < 3:
        raise ValueError(f"The number of links {len(links)} is less than 3.")

    for index, _ in enumerate(links[:-1]):
        if links[index + 1] not in [l for l, _ in links[index].links]:
            raise ValueError(f"The links are not a contiguous chain. {links[index + 1].name} is not connected to {links[index].name}.")

    for link in links[1:-1]:
        link.joint = Matrix()

    matrices = Matrix.slerp(links[0].get_joint_world(), links[-1].get_joint_world(), len(links))

    for link, matrix in zip(links, matrices):
        link.set_joint_world(matrix)
